#Making a PICO-8 TAS - Part 1: Intro

## Why

Speedrunners turn the process of beating a single-player video game into a
competitive activity, trying to finish games as quickly as humanly possible.
This has become increasingly mainstream (though still niche) in the past few
years with the rise in popularity of events like Games Done Quick.
However, there's another, smaller community interested in beating games quickly
but from a different perspective: how quickly can a game be beaten if humans
are taken out of the equation?

This is the question set out to be answered by a TAS, or Tool-Assisted Speedrun.

The goal is to create a sequence of inputs which complete the game
as quickly as possible, by any means necessary.
In practice, this means things like advancing the game frame-by-frame and
replaying small sections repeatedly to get them perfect.

In this series of posts I'm going to talk about my process of making a TAS for
[CELESTE Classic](https://mattmakesgames.itch.io/celeste) (referred to here as Celeste), one of my favourite
games.
This game was [run at Summer Games Done Quick 2016](https://www.youtube.com/watch?v=ffti-asjSeg), and I originally
released a [2:26 TAS of the game](https://www.youtube.com/watch?v=QVw6jHqCgbU) in Fall of last year.
For a couple reasons, I feel like it's time to update the TAS.
First, there's been a lot of new discoveries about the game, and I think it's
possible for a new TAS to save upwards of 8 seconds (quite a lot for a 2.5 minute game).
Second, there's been some changes to PICO-8 itself that will make the creation
of the TAS much easier, more pleasant, and more technically interesting.

## The Old TAS

The original TAS was made possible by [picolove](https://github.com/picolove/picolove), a PICO-8 emulator
written by [impbox](https://twitter.com/impbox).
I made some crude modifications to it that let it read in a sequence of inputs
from a text file and play them back.
That's how the entire TAS was made, by me entering button inputs into a text
file (that you can see [here](https://gist.github.com/justinj/7d7e149561fec0682a47)), then booting up the game and seeing if they worked.
It was extremely tedious to do it this way and I'm excited to create tools to
do it more easily.

## What is PICO-8

Celeste is a game for the [PICO-8](http://www.lexaloffle.com/pico-8.php), a fantasy video game console
created by Lexaloffle.
It emphasizes embracing restrictions to create interesting games and toys, in
particular, games are limited to 8192 Lua *tokens* of code.
When Celeste was first released, it used up nearly all of the available code
space, meaning adding extra mechanisms to control the game was effectively
impossible.
Since the release of Celeste, the way PICO-8 counts tokens has changed, giving
us around 1500 more tokens to work with, which is plenty to bolt on some tools.

Another recent addition to PICO-8 is GPIO pins.
This is a 128-byte chunk of memory usable for IO within PICO-8.
How it's exposed is dependent on platform (at the time of writing, it's not
exposed on the native client at all), but in the web player, it maps to a
global JavaScript array called `pico8_gpio`.

The combination of the increased code space and the GPIO pins means we can hook
up the PICO-8 web player to talk to JavaScript and then build the rest of the
TAS tooling from within JS (in theory).

My next post will talk about how I went about the process of hooking up the
GPIO pins to allow communication between JS-land and PICO-8-land. Thanks for reading!
