import React, { Component } from 'react';
import logo from './logo.svg';
import './App.css';
import Controller from './Controller';
import FrameList from './FrameList';

import { RPC, ArgTypes, connect } from './rpc';

function arrayEq(a, b) {
  if (a.length !== b.length) return false;
  for (let i = 0; i < a.length; i++) {
    if (a[i] !== b[i]) {
      return false;
    }
  }
  return true;
}

let loadCelesteLevel = RPC({
  id: 0,
  input: [
    ArgTypes.Byte
  ],
  output: []
});

let nextFrame = RPC({
  id: 1,
  input: [ ArgTypes.Byte ],
  output: [
    ArgTypes.Boolean, // isLevelCompleted
    ArgTypes.Number, // playerx
    ArgTypes.Number, // playery
    ArgTypes.Number, // playerxs
    ArgTypes.Number, // playerys
  ]
});

let dumpState = RPC({
  id: 2,
  input: [],
  output: [ ArgTypes.Unspecified ]
});

let pushState = RPC({
  id: 3,
  input: [ ArgTypes.Unspecified ],
  output: []
});

let showNumber = RPC({
  id: 4,
  input: [ ArgTypes.Number ],
  output: []
});

let playback = RPC({
  id: 5,
  input: [
    ArgTypes.Array(ArgTypes.Byte)
  ],
  output: []
});

let sendString = RPC({
  id: 6,
  input: [
    ArgTypes.String
  ],
  output: [
    ArgTypes.String
  ]
});

let bridge = connect();

class App extends Component {
  constructor() {
    super();
    this.storedStates = [];

    // this.state = {
    //   controller: 0,
    //   frames: [],
    //   currentFrame: -1,
    //   currentLevel: 0,
    //   currentFrames: 0,
    //   levels: (new Array(30).fill([[]]))
    // }
    this.state = JSON.parse(localStorage.getItem('saved-state'));
    this.state.currentLevel = 0;
    this.state.currentFrames = 0;
    this.state.currentFrame = -1;
    this.state.controller = 0;
    this.state.resyncing = false;

    this.handleChangeController = this.handleChangeController.bind(this);
  }

  componentDidMount() {
    this.keyListener = document.addEventListener('keydown', e => {
      if (e.keyCode === 'L'.charCodeAt(0)) {
        this.handleNext()
      } else if (e.keyCode === 'H'.charCodeAt(0)) {
        this.handleBack()
      } else if (e.keyCode === 37) {
        this.toggleButton(0)
      } else if (e.keyCode === 38) {
        this.toggleButton(2)
      } else if (e.keyCode === 39) {
        this.toggleButton(1)
      } else if (e.keyCode === 40) {
        this.toggleButton(3)
      } else if (e.keyCode === 'Z'.charCodeAt(0)) {
        this.toggleButton(4)
      } else if (e.keyCode === 'X'.charCodeAt(0)) {
        this.toggleButton(5)
      }
    })
  }

  toggleButton(which) {
    this.setState({
      controller: this.state.controller ^ Math.pow(2, which)
    });
  }

  render() {
    return (
      <div className="App">
        {this.state.resyncing && "RESYNCING"}
        <div style={{textAlign: 'center'}}>
          <div className='level-button' onClick={this.handleBack.bind(this)}>
            {'<'}
          </div>
          <div className='level-button' onClick={this.handleNext.bind(this)}>
            >
          </div>
        </div>
        <div style={{textAlign: 'center'}}>
          <Controller
            bitmask={this.state.controller}
            onChange={this.handleChangeController}
          />
        </div>
        <FrameList
          frames={this.currentFrames()}
          onPick={this.handlePickFrame.bind(this)}
          currentFrame={this.state.currentFrame}
        />
        <div style={{textAlign: 'center'}}>
          <div className='level-button' onClick={this.handlePlay.bind(this)}>
            Play back
          </div>
          {false && 'nvm this for now' && <div className='level-button' onClick={this.handleResync.bind(this)}>
            Resync
          </div>}
          <div className='level-button' onClick={this.handleClear.bind(this)}>
            Clear
          </div>
          <div className='level-button' onClick={this.handleBeginning.bind(this)}>
            Beginning
          </div>
          <div className='level-button' onClick={this.handleSave.bind(this)}>
            Save
          </div>
          <div className='level-button' onClick={this.handleLoad.bind(this)}>
            Load
          </div>
        </div>
        <div>
          <div>
            Level completed: {this.currentFrame().isDone ? 'yes' : 'no'}
          </div>
          <div>
          Curframe: {this.state.currentFrame}
          </div>
          <div>
          Player X: {this.currentFrame().playerx}
          </div>
          <div>
          Player Y: {this.currentFrame().playery}
          </div>
          <div>
          Player VX: {this.currentFrame().playerxs}
          </div>
          <div>
          Player VY: {this.currentFrame().playerys}
          </div>
        </div>
        <div>
          {this.state.levels[this.state.currentLevel].map((level, i) => {
            return <div className={'stored ' + (i === this.state.currentFrames ? 'picked' : '')} onClick={this.switchFrames(i)} />;
          })}
          <div className='copy' onClick={this.copy.bind(this)}>
            +
          </div>
        </div>
        <div>
          {(new Array(31).fill(0)).map((x, i) => {
            return <div className='level-button' onClick={this.handleClick.bind(this, i)}>
              {i+1}00m
            </div>
          })}
        </div>
        <div>
          {'{'}{this.currentFrames().map(f => f.controller).join(',')}{'}'}
        </div>
      </div>
    );
  }

  currentFrames() {
    return this.state.levels[this.state.currentLevel][this.state.currentFrames];
  }

  currentFrame() {
    return this.currentFrames()[this.state.currentFrame] || {};
  }

  handleResync() {
    this.setState({
      resyncing: true
    }, () => {
      let levels = this.state.levels.slice();
      let frames = levels[this.state.currentLevel].slice();
      levels[this.state.currentLevel] = frames;
      let inputs = frames[this.state.currentFrames()].map(x => x.controller);
      frames[this.state.currentFrames()] = [];
      this.setState({
        levels
      }, () => {
        this.resyncNextFrame(inputs);
      });
    });
  }

  resyncNextFrame(inputs) {
    if (inputs.length === 0) {
      this.setState({
        resyncing: false
      });
    } else {
      bridge.send(nextFrame(inputs[0])).then(() => {

      });
    }
  }

  handleSave() {
    localStorage.setItem('saved-state', JSON.stringify(this.state));
  }

  handleLoad() {
    this.setState(JSON.parse(localStorage.getItem('saved-state')));
  }

  switchFrames(i) {
    return () => {
      this.setState({
        currentFrames: i
      });
    }
  }

  handlePickFrame(i) {
    bridge.send(pushState(this.currentFrames()[i].state)).then(() => {
      console.log('the state we just sent is');
      console.log(JSON.stringify(this.currentFrames()[i].state));
      this.setState({
        currentFrame: i,
        controller: (this.currentFrames()[i + 1] && this.currentFrames()[i + 1].controller) || 0
      });
    });
  }

  copy() {
    let levels = this.state.levels.slice();
    let level = levels[this.state.currentLevel];
    level.push(this.currentFrames());
    levels[this.state.currentLevel] = level;
    this.setState({
      levels,
      currentFrames: level.length - 1
    });
  }

  handleBeginning() {
    this.setState({
      currentFrame: -1,
      controller: this.currentFrames()[0].controller
    }, () => {
      bridge.send(loadCelesteLevel(this.state.currentLevel));
    });
  }

  handleClear() {
    let levels = this.state.levels.slice();
    let frames = levels[this.state.currentLevel].slice();
    levels[this.state.currentLevel] = frames;
    let inputs = frames[this.state.currentFrames].map(x => x.controller);
    frames[this.state.currentFrames] = [];
    this.setState({
      levels,
      currentFrame: 0
    }, () => {
      bridge.send(loadCelesteLevel(this.state.currentLevel));
    });
  }

  handleClick(i, e) {
    e.preventDefault();
    bridge.send(loadCelesteLevel(i)).then(() => {
      this.setState({
        currentFrames: 0,
        currentFrame: this.state.levels[i][0].length - 1,
        currentLevel: i
      }, () => {
        let frames = this.currentFrames();
        if (frames.length > 0) {
          bridge.send(pushState(frames[frames.length - 1].state));
        }
      });
    });
  }

  handleChangeController(m) {
    this.setState({
      controller: m
    });
  }

  handlePush(e) {
    bridge.send(pushState(this.storedState));
  }

  handleDump(e) {
    bridge.send(dumpState()).then(([state]) => {
      this.storedState = state;
    });
  }

  handlePlay(e) {
    bridge.send(playback(this.currentFrames().map(f => f.controller)));
  }

  handleNext(e) {
    if (this.waitingOnNext) return;
    bridge.send(nextFrame(this.state.controller)).then((result) => {
      let [isDone, playerx, playery, playerxs, playerys] = result;
      this.waitingOnNext = true;
      bridge.send(dumpState()).then(([state]) => {
        console.log('the state we just got is');
        console.log(JSON.stringify(state));
        this.waitingOnNext = false;
        let controller = this.state.controller;
        let curFrame = this.currentFrames()[this.state.currentFrame + 1];
        let newFrameData = {
          state,
          controller,
          isDone,
          playerx,
          playery,
          playerxs,
          playerys,
        }
        let newFrames = this.currentFrames().slice(0, this.state.currentFrame + 1).concat([newFrameData]);
        let kept = false;
        if (curFrame && controller === curFrame.controller) {
          newFrames = newFrames.concat(this.currentFrames().slice(this.state.currentFrame + 2));
          kept = true;
        }
        let newLevels = this.updateFrames(newFrames);
        this.setState({
          frames: newFrames,
          currentFrame: this.state.currentFrame + 1,
          levels: newLevels,
          controller: (kept && this.currentFrames()[this.state.currentFrame + 2]) ? this.currentFrames()[this.state.currentFrame + 2].controller : this.state.controller
        })
      });
    });
  }

  updateFrames(newFrames) {
    let curLevel = this.state.levels[this.state.currentLevel].slice();
    curLevel[this.state.currentFrames] = newFrames;
    let levels = this.state.levels.slice();
    levels[this.state.currentLevel] = curLevel;
    return levels;
  }

  handleBack(e) {
    this.handlePickFrame(this.state.currentFrame - 1);
  }
}

export default App;
